/*
 * Linux Kernel Module for the Audine Camera
 * Copyright (C) 2000 Peter Kirchgessner
 * http://www.kirchgessner.net, mailto:peter@kirchgessner.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * The sample interface routines for the module have been taken from the
 * Linux Kernel Module Programming Guide by Ori Pomerantz contained
 * in the Linux Documentation Project.
 *
 */

static char *aud_versio = "v1.01";

/* Event history:
 * 1.01, 11-Jun-2000, Add more INQs
 * 1.00, 04-Jun-2000, Creation
 */

/* For tests not running as module, we must drop some stuff */
#ifndef NO_MODULE

/* Standard in kernel modules */
/* We're doing kernel work, specifically, a module */
#include <linux/kernel.h>
#include <linux/module.h>

/* Deal with CONFIG_MODVERSIONS */
#if CONFIG_MODVERSIONS==1
#define MODVERSIONS
#include <linux/modversions.h>
#endif

/* For character devices the definitions are here */
#include <linux/fs.h>

/* For transfer between kernel/user space */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
#include <asm/uaccess.h>
#endif

/* A wrapper which does next to nothing atpresent, but may */
/* * help for compatibility with future versions of Linux */
#include <linux/wrapper.h>

#endif

#include <sys/io.h>

/* In 2.2.3 /usr/include/linux/version.h includes a macro for this, */
/* but 2.0.35 doesn't - so I add it here if necessary. */
#ifndef KERNEL_VERSION
#define KERNEL_VERSION(a,b,c) ((a)*65536+(b)*256+(c))
#endif

/* Device Declarations **************************** */

/* The name for our device, as it will appear in /proc/devices */
#define DEVICE_NAME "audine"


/* The maximum length of the message buffer of the device */
#define BUF_LEN 4096

/* Is the device open right now? Used to prevent */
/* concurent access into the same device */
static int Device_Open = 0;


/* Informations about the CCD-Chip */

typedef struct {
 char name[32];
 int lead_rows, trail_rows;     /* Leading/trailing dummy rows */
 int lead_pixels, trail_pixels; /* Leading/trailing dummy pixels per row */
 int width, height;             /* Imageable area */
 int color;                     /* Color flag */
} CCD_INFO;

/* Entry 0 of the array is used as default for the audine */
static CCD_INFO ccds[] = {
 { "KAF0400", 4, 4, 14, 14, 768, 512, 0 },
 { "KAF0400C", 4, 4, 14, 14, 768, 512, 1 },
 { "KAF0401", 4, 4, 14, 14, 768, 512, 0 },
 { "KAF1600", 4, 4, 14, 14, 1536, 1024, 0 },
 { "KAF1602", 4, 4, 14, 14, 1536, 1024, 0 }
};

typedef struct {

 /* Communication buffers */

 char Message[BUF_LEN];  /* Message buffer (device -> user) */
 char *MessageNextFree;  /* Pointer to first free message byte */
 char *MessageLastValid; /* Pointer to last valid message byte */
 char *MessageNextOut;   /* Pointer to next byte to write */

 char Command[BUF_LEN];  /* Command buffer (user -> device) */
 char *CommandNextFree;  /* Pointer to next free command byte */
 char *CommandLastValid; /* End of command buffer */

 /* Camera state, configuration */

 CCD_INFO *ccd;          /* Informations about the CCD-chip */
 int DebugLevel;         /* Debug level */
 unsigned short Port;    /* Hardware port number */
 int h_binning, v_binning;  /* Horiz. and vert. binning */
 int ControlRegister;    /* The state of the output port control register */
 int AmplifierControl;   /* Bitmask to control amplifier */
 int ShutterControl;     /* Bitmask for shutter */
 int Aux0Control;        /* Bitmaks for auxiliary 0 */
 int Aux1Control;        /* Bitmaks for auxiliary 1 */

 int skip_lead_rows, skip_trail_rows;  /* Rows/columns to skip due to window */
 int skip_lead_pixels, skip_trail_pixels;

 int img_read;           /* Image read flag */
 int img_y;              /* current line to read */
} DeviceState;

typedef enum { AMP_CNTRL, SHUT_CNTRL, AUX0_CNTRL, AUX1_CNTRL } PORT_CNTRL;

/* Audines device state */
static DeviceState audstate;

#define DBG(a) ((a) <= as->DebugLevel)

#define ERR_NO_PORT "STATUS=1 (port not assigned)\n"
#define RETURN_IF_PORT_0(a) { if (((a)->Port) == 0) \
  { strcpy (status, ERR_NO_PORT); return;} }

#define RETURN_STATUS(s)  { strcpy (status, s); return; }

/* Declaration of local functions */
static void aud_command_handle (DeviceState *as);
static void aud_image_data_get (DeviceState *as);


static void
aud_state_init (DeviceState *as)

{
  memset (as, 0, sizeof (*as));

  as->MessageNextFree = as->MessageNextOut
    = as->Message;
  as->MessageLastValid = &(as->Message[BUF_LEN-1]);

  as->CommandNextFree = as->Command;
  as->CommandLastValid = &(as->Command[BUF_LEN-1]);

  as->ccd = &(ccds[0]);
  as->DebugLevel = 0;
  as->Port = 0;       /* Port not valid */

  as->h_binning = as->v_binning = 2;

  as->ControlRegister = 1;   /* The state of the output ports control */
                             /* register. (1 = amplifier on) */
  as->AmplifierControl = 0;  /* Dont handle Amplifier */
                             /* (therefore ControlRegister must be 1) */
  as->ShutterControl = 0;    /* Dont handle Shutter */
  as->Aux0Control = 0;       /* Dont handle Auxiliary 0 */
  as->Aux1Control = 0;       /* Dont handle Auxiliary 1 */

  as->skip_trail_rows = as->skip_lead_rows = as->skip_lead_pixels = as->skip_trail_pixels = 0;

  as->img_read = 0;   /* Not in image read mode */
  as->img_y = -1;     /* Current line to read */
}


/* This function is called whenever a process */
/* attempts to open the device file */

static int
device_open(struct inode *inode, struct file *file)

{
  DeviceState *as = &audstate;

  /* Dont use DBG() here. because variable as is not initialized */
#ifdef DEBUG
  printk ("audine device_open: inode=%p, file=%p\n", inode, file);

  /* This is how you get the minor device number in
   * case you have more than one physical device using
   * the driver. */
  printk("audine device_open: Device=%d.%d\n",
         inode->i_rdev >> 8, inode->i_rdev & 0xFF);
#endif

  /* We don't want to talk to two processes at the
   * same time */
  if (Device_Open)
    return -EBUSY;

  /* If this was a process, we would have had to be
   * more careful here.
   *
   * In the case of processes, the danger would be
   * that one process might have check Device_Open
   * and then be replaced by the scheduler by another
   * process which runs this function. Then, when the
   * first process was back on the CPU, it would assume
   * the device is still not open.
   *
   * However, Linux guarantees that a process won't be
   * replaced while it is running in kernel context.
   *
   * In the case of SMP, one CPU might increment
   * Device_Open while another CPU is here, right after
   * the check. However, in version 2.0 of the
   * kernel this is not a problem because there's a lock
   * to guarantee only one CPU will be kernel module at
   * the same time. This is bad in  terms of
   * performance, so version 2.2 changed it.
   * Unfortunately, I don't have access to an SMP box
   * to check how it works with SMP.
   */

  Device_Open++;

  /* Initialize structures */
  aud_state_init (as);

  /* Make sure that the module isn't removed while
   * the file is open by incrementing the usage count
   * (the number of opened references to the module, if
   * it's not zero rmmod will fail)
   */
  MOD_INC_USE_COUNT;

  return 0;
}


/* This function is called when a process closes the
 * device file. It doesn't have a return value in
 * version 2.0.x because it can't fail (you must ALWAYS
 * be able to close a device). In version 2.2.x it is
 * allowed to fail - but we won't let it.
 */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
static int
 device_release(struct inode *inode, struct file *file)
#else
static void
device_release(struct inode *inode, struct file *file)
#endif

{DeviceState *as = &audstate;

  if (DBG (8))
    printk ("audine device_release: inode=%p,file=%p\n", inode, file);

  /* We're now ready for our next caller */
  Device_Open --;

  /* Decrement the usage count, otherwise once you
   * opened the file you'll never get rid of the module.
   */
  MOD_DEC_USE_COUNT;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
  return 0;
#endif
}


/* This function is called whenever a process which
 * has already opened the device file attempts to
 * read from it. */

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
static ssize_t
device_read(struct file *file,
    char *buffer,    /* The buffer to fill with data */
    size_t length,   /* The length of the buffer */
    loff_t *offset)  /* Our offset in the file */
#else
static int
device_read(struct inode *inode,
    struct file *file,
    char *buffer,   /* The buffer to fill with
                     * the data */
    int length)     /* The length of the buffer
                     * (mustn't write beyond that!) */
#endif

{ DeviceState *as = &audstate;

  /* Number of bytes actually written to the buffer */
  int bytes_read = 0;

  for (;;)
  {
    /* When in image read mode, only get more image data */
    /* if buffer is empty. This makes sure that we can store */
    /* the scanline completely and at word boundary */
    if ((as->img_read) && (as->MessageNextFree == as->Message))
      aud_image_data_get (as);

    if (DBG (8))
      printk ("audine device_read: length=%d, av.=%d\n",
              length, as->MessageNextFree-as->MessageNextOut);

    /* If we're at the end of the message, return 0
     * (which signifies end of file) */
    if (as->MessageNextOut >= as->MessageNextFree)
      return 0;

    /* Actually put the data into the buffer */
    while (   length
           && (as->MessageNextOut < as->MessageNextFree))
    {

      /* Because the buffer is in the user data segment,
       * not the kernel data segment, assignment wouldn't
       * work. Instead, we have to use put_user which
       * copies data from the kernel data segment to the
       * user data segment. */
      put_user(*((as->MessageNextOut)++), buffer++);

      length --;
      bytes_read ++;
    }

    /* If nothing else in buffer, reset pointer to start of buffer */
    if (as->MessageNextOut == as->MessageNextFree)
      as->MessageNextOut = as->MessageNextFree = as->Message;

    /* If we are not in image read mode, we are finished */
    if (!as->img_read) break;

    /* Otherwise get a new scanline if we have room enough */
    if (length < as->ccd->width * 2) break;
  }

  if (DBG (8))
    printk ("audine: Read %d bytes, %d left\n", bytes_read, length);

   /* Read functions are supposed to return the number
    * of bytes actually inserted into the buffer */
  return bytes_read;
}



/* This function is called when somebody tries to write
 * into our device file */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
static ssize_t
device_write(struct file *file,
             const char *buffer,    /* The buffer */
             size_t length,   /* The length of the buffer */
             loff_t *offset)  /* Our offset in the file */
#else
static int
device_write(struct inode *inode,
             struct file *file,
             const char *buffer,
             int length)
#endif

{ DeviceState *as = &audstate;
  int i;

  if (DBG (8))
    printk ("audine device_write: file=%p, length=%d\n", file, length);

  for(i = 0; i < length && (as->CommandNextFree <= as->CommandLastValid); i++)
  {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
    get_user(*(as->CommandNextFree), buffer+i);
#else
    *(as->CommandNextFree) = get_user(buffer+i);
#endif
    if (*(as->CommandNextFree) == '\n')
    {
      *(as->CommandNextFree) = '\0';
      aud_command_handle (as);  /* Work up the command received */
                                /* Reset command buffer */
      as->CommandNextFree = as->Command;
    }
    else
    {
      (as->CommandNextFree)++;
    }
  }

  if (i < length)  /* Buffer overrun ? */
  {
    printk ("audine device_write: command buffer overrun\n");
    as->CommandNextFree = as->Command;
  }

  /* Again, return the number of input characters used */
  return i;
}


/* Module Declarations ***************************** */

/* The major device number for the device. This is
 * global (well, static, which in this context is global
 * within this file) because it has to be accessible
 * both for registration and for release. */
static int Major;

/* This structure will hold the functions to be
 * called when a process does something to the device
 * we created. Since a pointer to this structure is
 * kept in the devices table, it can't be local to
 * init_module. NULL is for unimplemented functions. */

struct file_operations Fops = {
  NULL,   /* seek */
  device_read,
  device_write,
  NULL,   /* readdir */
  NULL,   /* select */
  NULL,   /* ioctl */
  NULL,   /* mmap */
  device_open,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
  NULL,   /* flush */
#endif
  device_release  /* a.k.a. close */
};


/* Initialize the module - Register the character device */
int
init_module(void)

{
  /* Register the character device (atleast try) */
  Major = module_register_chrdev(0, DEVICE_NAME, &Fops);

  /* Negative values signify an error */
  if (Major < 0) {
    printk ("audine %s: %s device failed with %d\n", aud_versio,
            "Sorry, registering the character", Major);
    return Major;
  }

#ifdef NO_MODULE
  printk ("audine: Registration successful.\n");
#else
  printk ("audine %s: %s Major device number is %d.\n", aud_versio,
          "Registration successful.", Major);
  printk ("To talk to the device driver,\n");
  printk ("you'll have to create a device file:\n");
  printk ("mknod <name> c %d <minor>\n", Major);
  printk ("For more details see http://www.kirchgessner.net\n");
#endif

  return 0;
}

/* Cleanup - unregister the appropriate file from /proc */
void
cleanup_module(void)

{
  int ret;

  /* Unregister the device */
  ret = module_unregister_chrdev(Major, DEVICE_NAME);

  /* If there's an error, report it */
  if (ret < 0)
    printk("Error in unregister_chrdev: %d\n", ret);
}


/* The functions have been taken from the samples given on the */
/* Audine page astroccd.com/terre/audine/English/program.htm */

/********************************************************************/
/* Transfer one line of the image area into the horizontal register */
/********************************************************************/

void
ccd_hreg_load (register unsigned short base)

{
 outb (0xfb, base); /* 11111011b */
 outb (0xfb, base);
 outb (0xfb, base);
 outb (0xfb, base);
 outb (0xfb, base);
 outb (0xfb, base);
 outb (0xfb, base);
 outb (0xfb, base);

 outb (0xfa, base); /* 11111010b */
 outb (0xfa, base);
 outb (0xfa, base);
 outb (0xfa, base);
 outb (0xfa, base);
 outb (0xfa, base);
 outb (0xfa, base);
 outb (0xfa, base);

 outb (0xf9, base); /* 11111001b */
 outb (0xf9, base);
 outb (0xf9, base);
 outb (0xf9, base);
 outb (0xf9, base);
 outb (0xf9, base);
 outb (0xf9, base);
 outb (0xf9, base);

 outb (0xfa, base); /* 11111010b */
 outb (0xfa, base);
 outb (0xfa, base);
 outb (0xfa, base);
 outb (0xfa, base);
 outb (0xfa, base);
 outb (0xfa, base);
 outb (0xfa, base);

 outb (0xfb, base); /* 11111011b */
 outb (0xfb, base);
 outb (0xfb, base);
 outb (0xfb, base);
 outb (0xfb, base);
 outb (0xfb, base);
 outb (0xfb, base);
 outb (0xfb, base);
}

/***********************************************************************/
/* Shift out n pixel from the horizontal register without digitization */
/***********************************************************************/

void
ccd_pixels_read_fast (int n, register unsigned short base)

{
 while (n-- > 0)
 {
   outb (0xf7, base); /* 11110111b */
   outb (0xff, base); /* 11111111b */
   outb (0xfb, base); /* 11111011b */
 }
}

/********************************************************************/
/* Shift out complete horizontal register without digitization.     */
/* Used also to clear horizontal register.                          */
/********************************************************************/

static void
ccd_hreg_read_fast (const CCD_INFO *ccd, register unsigned short base)

{int j;

 j = ccd->lead_pixels + ccd->width + ccd->trail_pixels;
 ccd_pixels_read_fast (j, base);
}


/********************************************************************/
/* Shift out n lines of image area and clear horizontal register    */
/********************************************************************/

void
ccd_lines_read_fast (const CCD_INFO *ccd, unsigned short base, int n)

{int k;

 while (n > 0)
 {
   k = (n > 4) ? 4 : n;  /* To increase performance load 4 lines into hreg */
   while (k--)
   {
     ccd_hreg_load (base);
     n--;
   }
   ccd_hreg_read_fast (ccd, base);
 }
}


/********************************************************************/
/* Clear image area                                                 */
/********************************************************************/

static void
ccd_clear (const CCD_INFO *ccd, register unsigned short base)

{
 ccd_lines_read_fast (ccd, base, ccd->trail_rows + ccd->height + ccd->lead_rows);
}


/********************************************************************/
/* Read pixels from horizontal register and digitize                */
/********************************************************************/

static void
ccd_pixels_read (int npix, int binning, unsigned short *buf,
                 unsigned short port)

{unsigned short P, P2;
 int a1, a2, a3, a4;
 int b;

 P = port;
 P2 = P+1;

 while (npix-- > 0)
 {
   outb(247,P);   /* Reset */
   outb(255,P);   /* palier de reference  */
   outb(255,P);
   outb(255,P);
   outb(255,P);
   outb(255,P);
   outb(239,P);   /* clamp */

   for (b = 0; b < binning; b++)
   {
     outb(255,P);
     outb(251,P);   /* palier video */
   }

   outb(251,P);
   outb(251,P);
   outb(251,P);
   outb(251,P);
   outb(219,P);   /* start convert */
   outb(219,P);
   outb(219,P);
   outb(219,P);
   outb(219,P);
   outb(219,P);

   /* Read digitized values */
   a1 = (inb(P2)) >> 4;
   outb(91,P);
   a2 = (inb(P2)) >> 4;
   outb(155,P);
   a3 = (inb(P2)) >> 4;
   outb(27,P);
   a4 = (inb(P2)) >> 4;
   b = (a1+(a2 << 4)+(a3 << 8)+(a4 << 12))^0x8888;
   if (b > 32767) b = 32767;
   *(buf++) = (unsigned short)b;
 }
}


/********************************************************************/
/* Utility functions                                                */
/********************************************************************/

/* Convert a string into int, check for leading 0x. */
/* Next will point to the character behind the number (if not NULL) */
/* err is an error indicator (if not NULL) */
static int
strtoi (const char *s, const char **next, int *err)

{int i = 0;
 const char *sgn = NULL;
 int base = 10;

 if (err) *err = 0;
 while ((*s == ' ') || (*s == '\t')) s++;
 if (*s == '-')
 {
   if ((s[1] >= '0') && (s[1] <= '9'))
   {
     sgn = s;
     s++;
   }
   else
   {
     if (next) *next = s;
     if (err) *err = 1;
     return 0;
   }
 }
 else if ((*s < '0') || (*s > '9'))
 {
   if (next) *next = s;
   if (err) *err = 1;
   return 0;
 }
 /* Here is the first digit. Check which base we should use */
 if ((*s == '0') && (s[1] == 'x'))
 {
   if (sgn)  /* If we got a sign, we have now just 0 */
   {
     if (next) *next = s+1;
     return 0;
   }
   else if (   ((s[2] >= '0') && (s[2] <= '9'))
            || ((s[2] >= 'a') && (s[2] <= 'f'))
            || ((s[2] >= 'A') && (s[2] <= 'F')) )
   {
     base = 16;
     s += 2;
   }
 }
 if (base == 10)
 {
   while ((*s >= '0') && (*s <= '9'))
     i = 10*i + *(s++) - '0';
   if (sgn) i = -i;
 }
 else
 {
   for (;;)
   {
     if ((*s >= '0') && (*s <= '9'))
       i = 16*i + *(s++) - '0';
     else if ((*s >= 'a') && (*s <= 'f'))
       i = 16*i + *(s++) - 'a' + 10;
     else if ((*s >= 'A') && (*s <= 'F'))
       i = 16*i + *(s++) - 'A' + 10;
     else
       break;
   }
 }
 if (next) *next = s;
 return i;
}


/* Convert an int into a string. Result only valid until next call */
static char *
strfromi (int i)

{static char digit[32];
 char *ptr;
 int sgn = (i < 0);

 if (i == 0)
 {
   strcpy (digit, "0");
   return digit;
 }

 if (sgn) i = -i;

 ptr = digit + sizeof (digit) - 1;
 *ptr = '\0';
 while (i)
 {
   *(--ptr) = (i % 10) + '0';
   i /= 10;
 }
 if (sgn) *(--ptr) = '-';
 return ptr;
}


/* Read a word from a string and write it to sword. *next is the pointer */
/* to the next non white space character (if not NULL)*/
static void
strgetword (const char *s, char *sword, int sword_size,
            int delim, const char **next)

{int k;

 *sword = '\0';

 /* Skip leading whitespace */
 while ((*s == ' ') || (*s == '\t')) s++;
 if ((*s == '\0') || (*s == delim))
 {
   if (next) *next = s;
   return;
 }

 k = 0;
 for (;;)   /* Copy the word */
 {
   *(sword++) = *(s++);
   k++;
   if (k+1 == sword_size) break; /* No more room ? */
   if ((*s == ' ') || (*s == '\t') || (*s == '\0') || (*s == delim))
     break;

 }
 *sword = '\0';
 while ((*s == ' ') || (*s == '\t')) s++;
 if (next) *next = s;
}


static
void aud_control_register_setbit (DeviceState *as, int value)

{
 as->ControlRegister |= value;
}


static
void aud_control_register_clearbit (DeviceState *as, int value)

{
 as->ControlRegister &= ( (~value) & 0xff);
}


static
void aud_control_register_write (DeviceState *as)

{int value = as->ControlRegister;

 /* The samples from the Audine site say to switch on amplifier */
 /* one must write the value 1 to the control register. */
 /* But it seems that we have to write 0. So we invert bits 1 and 3. */
 value ^= 5;

 if (as->Port)
   outb (value, as->Port + 2);
}


/* Write a message string to the message buffer */
static void
aud_message_write (DeviceState *as, char *s)

{
 while ((*s) && (as->MessageNextFree <= as->MessageLastValid))
   *((as->MessageNextFree)++) = *(s++);
}


/********************************************************************/
/* Read one line from the camera. The data can be stored at the     */
/* start of the Message buffer.                                     */
/********************************************************************/

static void
aud_image_data_get (DeviceState *as)

{CCD_INFO *ccd = as->ccd;
 unsigned short port = as->Port;
 int y, k, npix;
 int xmin, ymin, xmax, ymax;

 xmin = as->skip_lead_pixels;
 xmax = ccd->width - as->skip_trail_pixels - 1;
 ymin = as->skip_lead_rows;
 ymax = ccd->height - as->skip_trail_rows - 1;
 npix = (xmax-xmin+1)/as->h_binning;
 
 y = as->img_y;

 if (y < 0)   /* leading rows to skip ? */
 {
   ccd_lines_read_fast (ccd, port, ccd->lead_rows + as->skip_lead_rows);
   y = as->skip_lead_rows;  /* is ymin */
 }

 /* XXXXX This works for multi read */
 /* ccd_pixels_read (ccd->width+ccd->lead_pixels+ccd->trail_pixels,    */
 /*                    1, (unsigned short *)as->Message, port);  */

 if (y+as->v_binning-1 <= ymax)  /* A line to read ? */
 {
   ccd_hreg_read_fast (ccd, port);  /* Clear horizontal register */

   k = as->v_binning;    /* Due to binning load horizontal register */
   while (k-- > 0)
   {
     ccd_hreg_load (port);
     y++;
   }

   /* Skip first pixels */
   k = ccd->lead_pixels + as->skip_lead_pixels;
   ccd_pixels_read_fast (k, port);

   /* Get real image data */
   ccd_pixels_read (npix, as->h_binning, (unsigned short *)as->Message, port);
   as->MessageNextFree = as->Message + npix*2;

   /* Skip last dummy pixels */
   k = (ccd->width + ccd->trail_pixels) - (as->skip_lead_pixels + npix*as->h_binning);
   ccd_pixels_read_fast (k, port);
 }

 if (y+as->v_binning-1 <= ymax) /* Still something to read ? */
 {
   as->img_y = y;  /* Save line */
 }
 else   /* Skip trailing rows and end image read mode */
 {
   k = ccd->height+ccd->trail_rows-y;
   ccd_lines_read_fast (ccd, port, k);

   as->img_read = 0;
   as->img_y = -1;
 }
}


/********************************************************************/
/* Functions to handle the various commands. They receive the state */
/* of the device and the parameter from the command. If a message   */
/* is to generate, it is returned in status. Status is at least 80  */
/* characters wide and initialized with the empty string.           */
/********************************************************************/

static void
aud_amplifier_handler (DeviceState *as, const char *param, char *status)

{char cmd[32];

 RETURN_IF_PORT_0 (as);
 if (as->AmplifierControl == 0)
   RETURN_STATUS ("STATUS=1 (no amplifier control)\n");

 strgetword (param, cmd, sizeof (cmd), 0, NULL);

 if (strcmp (cmd, "ON") == 0)
   aud_control_register_setbit (as, as->AmplifierControl);
 else if (strcmp (cmd, "OFF") == 0)
   aud_control_register_clearbit (as, as->AmplifierControl);
 else
   RETURN_STATUS ("STATUS=1 (parameter must be ON or OFF)\n");

 aud_control_register_write (as);
}


static void
aud_aux0_handler (DeviceState *as, const char *param, char *status)

{char cmd[32];

 RETURN_IF_PORT_0 (as);
 if (as->Aux0Control == 0)
   RETURN_STATUS ("STATUS=1 (no aux0 control)\n");

 strgetword (param, cmd, sizeof (cmd), 0, NULL);

 if (strcmp (cmd, "ON") == 0)
   aud_control_register_setbit (as, as->Aux0Control);
 else if (strcmp (cmd, "OFF") == 0)
   aud_control_register_clearbit (as, as->Aux0Control);
 else
   RETURN_STATUS ("STATUS=1 (parameter must be ON or OFF)\n");

 aud_control_register_write (as);
}


static void
aud_aux1_handler (DeviceState *as, const char *param, char *status)

{char cmd[32];

 RETURN_IF_PORT_0 (as);
 if (as->Aux1Control == 0)
   RETURN_STATUS ("STATUS=1 (no aux1 control)\n");

 strgetword (param, cmd, sizeof (cmd), 0, NULL);

 if (strcmp (cmd, "ON") == 0)
   aud_control_register_setbit (as, as->Aux1Control);
 else if (strcmp (cmd, "OFF") == 0)
   aud_control_register_clearbit (as, as->Aux1Control);
 else
   RETURN_STATUS ("STATUS=1 (parameter must be ON or OFF)\n");

 aud_control_register_write (as);
}


static void
aud_shutter_handler (DeviceState *as, const char *param, char *status)

{char cmd[32];

 RETURN_IF_PORT_0 (as);
 if (as->ShutterControl == 0)
   RETURN_STATUS ("STATUS=1 (no shutter control)\n");

 strgetword (param, cmd, sizeof (cmd), 0, NULL);

 if (strcmp (cmd, "ON") == 0)
   aud_control_register_setbit (as, as->ShutterControl);
 else if (strcmp (cmd, "OFF") == 0)
   aud_control_register_clearbit (as, as->ShutterControl);
 else
   RETURN_STATUS ("STATUS=1 (parameter must be ON or OFF)\n");

 aud_control_register_write (as);
}


static void
aud_xxx_control_handler (DeviceState *as, const char *param, char *status,
                         PORT_CNTRL control_type)

{int ctrl, err;

 ctrl = strtoi (param, NULL, &err);
 if (err)
 {
   strcpy (status, "STATUS=1 (non numeric value)\n");
   return;
 }
 if ((ctrl != 0) && (ctrl != 1) && (ctrl != 2) && (ctrl != 4) && (ctrl != 8))
 {
   strcpy (status, "STATUS=1 (invalid value)\n");
   return;
 }
 switch (control_type)
 {
   case AMP_CNTRL: as->AmplifierControl = ctrl; break;
   case SHUT_CNTRL: as->ShutterControl = ctrl; break;
   case AUX0_CNTRL: as->Aux0Control = ctrl; break;
   case AUX1_CNTRL: as->Aux1Control = ctrl; break;
 }
}

static void
aud_amp_control_handler (DeviceState *as, const char *param, char *status)

{
 aud_xxx_control_handler (as, param, status, AMP_CNTRL);
}


static void
aud_aux0_control_handler (DeviceState *as, const char *param, char *status)

{
 aud_xxx_control_handler (as, param, status, AUX0_CNTRL);
}


static void
aud_aux1_control_handler (DeviceState *as, const char *param, char *status)

{
 aud_xxx_control_handler (as, param, status, AUX1_CNTRL);
}


static void
aud_shut_control_handler (DeviceState *as, const char *param, char *status)

{
 aud_xxx_control_handler (as, param, status, SHUT_CNTRL);
}


static void
aud_binning_handler (DeviceState *as, const char *param, char *status)

{int hb, vb;

 hb = param[0] - '0';
 vb = param[2] - '0';
 if (   (hb >= 1) && (hb <= 5) && (vb >=1) && (vb <= 5)
     && ((param[1] == 'x') || (param[1] == 'X')))
 {
   as->h_binning = hb;
   as->v_binning = vb;
 }
 else
 {
   strcpy (status, "STATUS=1 (invalid binning)\n");
 }
}


static void
aud_ccd_handler (DeviceState *as, const char *param, char *status)

{char cmd[32];
 int k, nccds;

 strgetword (param, cmd, sizeof (cmd), 0, NULL);
 if (strlen (cmd) == 0)
   RETURN_STATUS ("STATUS=1 (no ccd specified)\n");

 nccds = sizeof (ccds) / sizeof (ccds[0]);
 for (k = 0; k < nccds; k++)
 {
   if (strcmp (cmd, ccds[k].name) == 0)
   {
     as->ccd = &(ccds[k]);

     /* Reset window */
     as->skip_lead_pixels = as->skip_trail_pixels = as->skip_lead_rows = as->skip_trail_rows = 0;

     return;
   }
 }
 strcpy (status, "STATUS=1 (invalid ccd)\n");
}


static void
aud_clear_handler (DeviceState *as, const char *param, char *status)

{int numclear;

 RETURN_IF_PORT_0 (as);

 numclear = strtoi (param, NULL, NULL);
 if (numclear < 1) numclear = 1;
 else if (numclear > 4) numclear = 4;

 if (DBG(8))
   printk ("aud_clear_handle: Clear %d times\n", numclear);

 while (numclear-- > 0)
   ccd_clear (as->ccd, as->Port);
}


static void
aud_inq_handler (DeviceState *as, const char *param, char *status)

{char inq[32];
 const char *next;
 CCD_INFO *ccd = as->ccd;

 strgetword (param, inq, sizeof (inq), 0, &next);
 if (strcmp (inq, "CCD") == 0)
 {
   strcpy (status, "CCD=");
   strcat (status, ccd->name);
   strcat (status, "\n");
 }
 else if (strcmp (inq, "CCDSIZE") == 0)
 {
   strcpy (status, "CCDSIZE=");
   strcat (status, strfromi (ccd->width)); strcat (status, " ");
   strcat (status, strfromi (ccd->height));
   strcat (status, "\n");
 }
 else if (strcmp (inq, "IMAGE") == 0)
 {
   RETURN_IF_PORT_0 (as);

   as->img_read = 1;  /* Activate image read mode */
   as->img_y = -1;
 }
 else if (strcmp (inq, "GEOMETRY") == 0)
 {int width, height;

   width = ccd->width - as->skip_lead_pixels - as->skip_trail_pixels;
   height = ccd->height - as->skip_lead_rows - as->skip_trail_rows;
   strcpy (status, "GEOMETRY=");
   strcat (status, strfromi (width/as->h_binning)); strcat (status, " ");
   strcat (status, strfromi (height/as->v_binning)); strcat (status, " ");
   strcat (status, strfromi (2)); strcat (status, " ");
   strcat (status, strfromi (32767)); strcat (status, " ");
   strcat (status, strfromi (ccd->color));
   strcat (status, "\n");
 }
 else
 {
   strcpy (status, "STATUS=1 (invalid inq)\n");
 }
}


static void
aud_port_handler (DeviceState *as, const char *param, char *status)

{int port, err;

 port = strtoi (param, NULL, &err);
 if (err)
 {
   strcpy (status, "STATUS=1 (invalid port number)\n");
 }
 else if ((port < 0x278) || (port > 0x778))
 {
   strcpy (status, "STATUS=1 (port number out of range)\n");
 }
 else
 {
   as->Port = (unsigned short)port;
   aud_control_register_write (as);
 }
}


static void
aud_window_handler (DeviceState *as, const char *param, char *status)

{int xmin, ymin, width, height, err;
 const char *start, *next;
 CCD_INFO *ccd = as->ccd;

 start = param;
 xmin = strtoi (param, &next, &err);
 if (err)
 {
   RETURN_STATUS ("STATUS=1 (invalid xmin)\n");
 }
 start = next;
 ymin = strtoi (start, &next, &err);
 if (err)
 {
   RETURN_STATUS ("STATUS=1 (invalid ymin)\n");
 }
 start = next;
 width = strtoi (start, &next, &err);
 if (err)
 {
   RETURN_STATUS ("STATUS=1 (invalid width)\n");
 }
 start = next;
 height = strtoi (start, &next, &err);
 if (err)
 {
   RETURN_STATUS ("STATUS=1 (invalid height)\n");
 }
 if (xmin < 0) xmin = 0;
 else if (xmin >= ccd->width) xmin = ccd->width-1;
 if (ymin < 0) ymin = 0;
 else if (ymin >= ccd->height) ymin = ccd->height-1;
 if (xmin + width > ccd->width) width = ccd->width - xmin;
 if (ymin + height > ccd->height) height = ccd->height - ymin;

 as->skip_lead_pixels = xmin;
 as->skip_trail_pixels = ccd->width-(xmin+width);
 as->skip_lead_rows = ymin;
 as->skip_trail_rows = ccd->height-(ymin+height);
}


/* An array that keeps all the pointers to the handle functions */
/* along with the command string they can work on */

static struct handler_s {
  char *command;
  void (*handlefunc) (DeviceState *as, const char *param, char *status);
} aud_cmd_handlers[] = {

  { "AMPLIFIER", aud_amplifier_handler },
  { "AMPLIFIERCONTROL", aud_amp_control_handler },
  { "AUX0", aud_aux0_handler },
  { "AUX1", aud_aux1_handler },
  { "AUX0CONTROL", aud_aux0_control_handler },
  { "AUX1CONTROL", aud_aux1_control_handler },
  { "BINNING", aud_binning_handler },
  { "CCD",     aud_ccd_handler },
  { "CLEAR",   aud_clear_handler },
  { "INQ",     aud_inq_handler },
  { "PORT",    aud_port_handler },
  { "SHUTTER", aud_shutter_handler },
  { "SHUTTERCONTROL", aud_shut_control_handler },
  { "WINDOW",  aud_window_handler }
};


/* Work on the command received in as->Command. */
/* The command is null-terminated */

static void
aud_command_handle (DeviceState *as)

{char cmd[32], status[80];
 const char *param;
 int k, nhandlers;

 strgetword (as->Command, cmd, sizeof (cmd), '=', &param);
 if (cmd[0] == '#') return;  /* Comment */
 if (*param != '=')
 {
   aud_message_write (as, "STATUS=1 (Invalid syntax, cmd=param)\n");
   return;
 }

 /* *param is '='. Search first non white space character in param */
 do
 {
   param++;
 } while ((*param == ' ') || (*param == '\t'));

 status[0] = '\0';
 nhandlers = sizeof (aud_cmd_handlers)/sizeof (aud_cmd_handlers[0]);
 for (k = 0; k < nhandlers; k++)
 {
   if (strcmp (aud_cmd_handlers[k].command, cmd) == 0)
   {
     (*aud_cmd_handlers[k].handlefunc) (as, param, status);
     if (status[0] != '\0')
       aud_message_write (as, status);
     break;
   }
 }
 if (k >= nhandlers)
   aud_message_write (as, "STATUS=1 (Unknown command)\n");
}
