/*
 * Test program for the audine camera
 * Copyright (C) 2000 Peter Kirchgessner
 * http://www.kirchgessner.net, mailto:peter@kirchgessner.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 *
 */

static char *rdaud_versio = "v1.01";

/* Event history:
 * 1.01, 11-Jun-2000, Add option -s
 * 1.00, 04-Jun-2000, Creation
 */

#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <time.h>

/* If we dont want to use the module, redefine some stuff */

#ifdef NO_MODULE

/* Some stuff to make the module code running in user mode */
static struct inode { int i_rdev; } inode_var = { 0 };
static struct file { char *dummy; } file_var = { NULL };
struct file_operations { void *d0,*d1,*d2,*d3,*d4,*d5,*d6,*d7,*d8,*d9; };
#define printk printf
#define EBUSY  16
#define MOD_INC_USE_COUNT /* */
#define MOD_DEC_USE_COUNT /* */
#define put_user(k,u) *(u) = (k)
#define get_user(u) (*(u))
#define module_register_chrdev(a,b,c) (0)
#define module_unregister_chrdev(a,b) (0)

/* Get modules code */
#include "audine.c"

#else

static int aud_fd = -1;  /* The file descriptor for the audine camera */

#endif

#define DEVICE_FILE "/dev/audine"

/* Time in ms to activate amplifier before use */
#define AMPLIFIER_DELAY 300

/* Time in ms to activate shutter before/after exposure */
#define SHUTTER_DELAY 200

/* Test purposes. Read the image with a single read() or multiple read()s */
static int single_read = 1;

static void
send_audine_cmd (char *cmd)

{int k, numbytes = strlen (cmd);

#ifdef NO_MODULE
 k = device_write (&inode_var, &file_var, cmd, strlen(cmd));
#else
 k = write (aud_fd, cmd, numbytes);
#endif

 if (k != numbytes)
  fprintf (stderr, "send_audine_cmd: Could only write %d of %d bytes.\n",
           k, numbytes);
}


static void
read_audine_answer (char *answer)

{char buf[128], *ptr;
 int len;

 ptr = buf;
 for (;;)
 {
#ifdef NO_MODULE
   len = device_read (&inode_var, &file_var, ptr, sizeof (buf) - (ptr-buf));
#else
   len = read (aud_fd, ptr, sizeof (buf) - (ptr-buf));
#endif
   if (len <= 0) break;
   ptr += len;
 }
 len = ptr-buf;
 fwrite (buf, 1, len, stdout);
 fflush (stdout);
 if (answer)
 {
   memcpy (answer, buf, len);
   answer[len] = '\0';
 }
}


static void
read_audine_data (char *buf, int width, int height)

{int len, nread;
 int nbytes = width * height * 2;

 /* Here we could decide if we want to have the image with one read */
 /* or with multiple reads. A single read would block the system */
 /* for several seconds (18 seconds for 768x512). Multiple reads */
 /* wouldnt block. But multiple reads gives problems with the timing */
 /* in the audine and destroy the image. Thus we block the system. */
 while (nbytes > 0)
 {
   if (single_read)
   {
     nread = nbytes;
   }
   else
   {
     nread = width*2;
     if (nread > nbytes) nread = nbytes;
   }
#ifdef NO_MODULE
   len = device_read (&inode_var, &file_var, buf, nread);
#else
   len = read (aud_fd, buf, nread);
#endif
   if (len <= 0)
   {
     fprintf (stderr, "read_audine_data: EOF\n");
     return;
   }
   nbytes -= len;
   buf += len;
 }
}


#define FITS_WRITE_BOOLCARD(fp,key,value) \
{char card[81]; \
 sprintf (card, "%-8.8s= %20s%50s", key, value ? "T" : "F", " "); \
 fwrite (card, 1, 80, fp); }

#define FITS_WRITE_LONGCARD(fp,key,value) \
{char card[81]; \
 sprintf (card, "%-8.8s= %20ld%50s", key, (long)value, " "); \
 fwrite (card, 1, 80, fp); }

#define FITS_WRITE_DOUBLECARD(fp,key,value) \
{char card[81], dbl[21], *istr; \
 sprintf (dbl, "%20f", (double)value); istr = strstr (dbl, "e"); \
 if (istr) *istr = 'E'; \
 sprintf (card, "%-8.8s= %20.20s%50s", key, dbl, " "); \
 fwrite (card, 1, 80, fp); }

#define FITS_WRITE_STRINGCARD(fp,key,value) \
{char card[81]; int k;\
 sprintf (card, "%-8.8s= \'%s", key, value); \
 for (k = strlen (card); k < 81; k++) card[k] = ' '; \
 k = strlen (key); if (k < 8) card[19] = '\''; else card[11+k] = '\''; \
 fwrite (card, 1, 80, fp); }

#define FITS_WRITE_CARD(fp,value) \
{char card[81]; \
 sprintf (card, "%-80.80s", value); \
 fwrite (card, 1, 80, fp); }

static void
write_fits (int width, int height, char *img, int nbytes)

{FILE *fp;
 char value[50];
 unsigned short *us_img = (unsigned short *)img;
 unsigned short *row;
 int x, y, high, low;
 long pos;
 time_t timp;

 fp = fopen ("audine.fits", "w");
 if (fp)
 {
   FITS_WRITE_BOOLCARD (fp, "SIMPLE", 1);
   FITS_WRITE_LONGCARD (fp, "BITPIX", 16);
   FITS_WRITE_LONGCARD (fp, "NAXIS", 2);
   FITS_WRITE_LONGCARD (fp, "NAXIS1", width);
   FITS_WRITE_LONGCARD (fp, "NAXIS2", height);
   FITS_WRITE_DOUBLECARD (fp, "BZERO", 0.0);
   FITS_WRITE_DOUBLECARD (fp, "BSCALE", 1.0);
   FITS_WRITE_DOUBLECARD (fp, "DATAMIN", 0.0);
   FITS_WRITE_DOUBLECARD (fp, "DATAMAX", 32767.0);
   FITS_WRITE_CARD (fp, " ");
   FITS_WRITE_CARD (fp, "HISTORY THIS FILE WAS GENERATED BY RDAUD");
   FITS_WRITE_CARD (fp, " ");

   timp = time (NULL);
   strftime (value, sizeof (value), "%d/%m/%y", gmtime (&timp));
   FITS_WRITE_STRINGCARD (fp, "DATE", value);

   FITS_WRITE_CARD (fp, "END");

   /* Fill up primary HDU to multiple of 2880 bytes */
   fflush (fp);
   pos = ftell (fp) % 2880;
   if (pos != 0)
   {
     pos = 2880 - pos;
     while (pos-- > 0)
       putc (' ', fp);
   }

   /* FITS standard requires integer data to be most significant byte first, */
   /* image origin bottom left. We want to create an astronomical oriented */
   /* image (top is bottom, left is right) */
   for (y = 0; y < height; y++)
   {
     row = us_img + y*width;
     for (x = width-1; x >= 0; x--)
     {
       high = (row[x] >> 8) & 0xff;
       low = (row[x] & 0xff);
       putc (high, fp);
       putc (low, fp);
     }
   }
   /* Fill up file to multiple of 2880 bytes */
   fflush (fp);
   pos = ftell (fp) % 2880;
   if (pos != 0)
   {
     pos = 2880 - pos;
     while (pos-- > 0)
       putc (0, fp);
   }
   fclose (fp);
 }
 else
 {
   fprintf (stderr, "Could not open audine.fits for writing\n");
 }
}


static void
write_pgm (int width, int height, char *img, int nbytes)

{FILE *fp;

 fp = fopen ("audine.pgm", "w");
 if (fp)
 {
   fprintf (fp, "P5\n%d %d\n32767\n", width, height);
   fwrite (img, 1, nbytes, fp);
   fclose (fp);
 }
 else
   fprintf (stderr, "Could not open audine.pgm for writing\n");
}


static void
auto_threshhold (unsigned short *img, int npix, double dthreshhold)

{int count[32768];
 int k, sum, newval;
 int minval, maxval;
 int threshhold;

 threshhold = (npix * dthreshhold) / 100.0;

 memset (count, 0, sizeof (count));
 for (k = 0; k < npix; k++)
 {
   (count[img[k]])++;
 }
 minval = 0;
 sum = 0;
 for (k = 0; k < 32768; k++)
 {
   sum += count[k];
   if (sum < threshhold) minval = k;
   else break;
 }
 maxval = 32767;
 sum = 0;
 for (k = 32767; k >= 0; k--)
 {
   sum += count[k];
   if (sum < threshhold) maxval = k;
   else break;
 }
printf ("minval/maxval=%d/%d\n", minval, maxval);
 if ((minval == 0) && (maxval == 32767)) return;
 if (minval >= maxval) return;

 for (k = 0; k < npix; k++)
 {
   newval = (((int)img[k]) - minval);
   if (newval <= 0) img[k] = 0;
   else
   {
     newval = (newval * 32767) / (maxval - minval);
     if (newval >= 32767) img[k] = 32767;
     else img[k] = (unsigned short)newval;
   }
 }
}


static void
usage_exit (const char *prog)

{
 fprintf (stderr, "Usage:\n");
 fprintf (stderr, "%s PORT=<port> [<audine-cmd>] [-t <time>] [-v]\n", prog);
 exit (0);
}

int
main (int argc, char *argv[])

{char command[128], *img;
 int width = 0, height = 0;
 int nbytes, k;
 int use_amplifier = 0;
 int use_shutter = 0;
 int exposure = 0;
 int fits = 1;
 double threshhold = 0;
#ifdef NO_MODULE
 char *ep;
#endif

 if ((argc <= 1) || (strncmp (argv[1], "-h", 2) == 0))
   usage_exit (argv[0]);

#ifdef NO_MODULE

 init_module ();
 device_open (&inode_var, &file_var);

#else

 aud_fd = open (DEVICE_FILE, O_RDWR);
 if (aud_fd < 0)
 {
   fprintf (stderr, "Error opening %s\n", DEVICE_FILE);
   exit (1);
 }
#endif

 for (k = 1; k < argc; k++)  /* Process command line options */
 {
   if (strchr (argv[k], '=') != NULL)  /* Audine command ? */
   {
     use_shutter |= (strncmp (argv[k], "SHUTTERCONTROL=", 15) == 0);
     use_amplifier |= (strncmp (argv[k], "AMPLIFIERCONTROL=", 16) == 0);
#ifdef NO_MODULE
     if (strncmp (argv[k], "PORT=", 5) == 0)
     {int port;

       port = (int)strtol (argv[k]+5, &ep, 0);
       if (ioperm (port, 3, 1) != 0)
       {
          fprintf (stderr, "Cant get permission for port 0x%x\n", port);
          exit (1);
       }
     }
#endif
     sprintf (command, "%s\n", argv[k]);
     send_audine_cmd (command);
     read_audine_answer (command);
   }
   else if (strcmp (argv[k], "-v") == 0)
   {
     printf ("Testprogram %s for Audine interface\n", rdaud_versio);
   }
   else if ((strcmp (argv[k], "-s") == 0) && (k+1 < argc))
   {
     single_read = (argv[++k][0] != '0');
   }
   else if ((strcmp (argv[k], "-a") == 0) && (k+1 < argc))
   {
     if (sscanf (argv[k+1], "%lf", &threshhold) != 1)
       threshhold = 0.0;
     else if (threshhold < 0.0) threshhold = 0.0;
     else if (threshhold > 100.0) threshhold = 100.0;
     k++;
   }
   else if ((strcmp (argv[k], "-t") == 0) && (k+1 < argc))
   {double dex = 0.0;

     if (sscanf (argv[k+1], "%lf", &dex) != 1)
       dex = 0.0;
     else if (dex < 0.0) dex = 0.0;
     exposure = (int)(dex * 1000.0); /* Get time in ms */
     k++;
   }
   else if (strcmp (argv[k], "-pgm") == 0)
   {
     fits = 0;
   }
   else
   {
     usage_exit (argv[0]);
   }
 }

 /* Get geometry of image */
 send_audine_cmd ("INQ=GEOMETRY\n");
 read_audine_answer (command);
 if (   (strncmp (command, "GEOMETRY=", 9) != 0)
     || (sscanf (command+9, "%d%d", &width, &height) != 2))
 {
   fprintf (stderr, "Could not inquire image geometry\n");
 }
 else
 {
   nbytes = width*height*2;
   img = malloc (nbytes);
   if (img == NULL)
   {
     fprintf (stderr, "Could not malloc picture\n");
   }
   else
   {
     /* Now we are in the exposure phase */
     if (exposure <= 1) use_amplifier = 0;

     if (use_amplifier)
     {
       send_audine_cmd ("AMPLIFIER=OFF\n");
       read_audine_answer (command);
     }
     if (use_shutter)
     {
       send_audine_cmd ("SHUTTER=ON\n");
       read_audine_answer (command);
       usleep (SHUTTER_DELAY * 1000); /* Wait until shutter is open */
     }

     /* Now we have to wait for the exposure time. Activate amplifier */
     /* and or shutter in time. */
     if (use_shutter && use_amplifier)
     {
       if (SHUTTER_DELAY < AMPLIFIER_DELAY)
       {
         usleep ((exposure - AMPLIFIER_DELAY) * 1000);
         send_audine_cmd ("AMPLIFIER=ON\n");
         read_audine_answer (command);
         usleep ((AMPLIFIER_DELAY - SHUTTER_DELAY) * 1000);
         send_audine_cmd ("SHUTTER=OFF\n");
         read_audine_answer (command);
         usleep (SHUTTER_DELAY * 1000);
       }
       else
       {
         usleep ((exposure - SHUTTER_DELAY) * 1000);
         send_audine_cmd ("SHUTTER=OFF\n");
         read_audine_answer (command);
         usleep ((SHUTTER_DELAY - AMPLIFIER_DELAY) * 1000);
         send_audine_cmd ("AMPLIFIER=ON\n");
         read_audine_answer (command);
         usleep (AMPLIFIER_DELAY * 1000);
       }
     }
     else if (use_amplifier) /* Solve use of amplifier */
     {
       usleep ((exposure - AMPLIFIER_DELAY) * 1000);
       send_audine_cmd ("AMPLIFIER=ON\n");
       read_audine_answer (command);
       usleep (AMPLIFIER_DELAY * 1000);
     }
     else if (use_shutter) /* Sole use of shutter */
     {
       usleep ((exposure - SHUTTER_DELAY) * 1000);
       send_audine_cmd ("SHUTTER=OFF\n");
       read_audine_answer (command);
       usleep (SHUTTER_DELAY * 1000);
     }
     else
     {
       usleep (exposure * 1000);
     }

     /* Now we can get image data */
     printf ("*** Start reading image ***\n");
     fflush (stdout);
     send_audine_cmd ("INQ=IMAGE\n");
     read_audine_data (img, width, height);
     printf ("*** Finished reading image ***\n");
     fflush (stdout);
     if (threshhold > 0.0)
       auto_threshhold ((unsigned short *)img, width*height, threshhold);
     if (fits)
       write_fits (width, height, img, nbytes);
     else
       write_pgm (width, height, img, nbytes);
     free (img);
   }
 }

#ifdef NO_MODULE
 device_release (&inode_var, &file_var);
 cleanup_module ();
#else
 close (aud_fd);
#endif

 return 0;
}
